#include <windows.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <math.h>
#include <stdlib.h>

typedef struct
{
   WAVEFORMATEX wfx;
   WORD wID;
   DWORD fdwFlags;
   WORD nBlockSize;
   WORD nFramesPerBlock;
   WORD nCodecDelay;
}MPEGLAYER3WAVEFORMAT;
#define MPEGLAYER3_WFX_EXTRA_BYTES 12
#define  WAVE_FORMAT_MPEGLAYER3 0x0055

#define SAMPLE_RATE 22050
#define MP3_BLOCK_SIZE 522
#define MPEGLAYER3_FLAG_PADDING_OFF 0x00000002

void CALLBACK CallBackProc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
  if(uMsg == WOM_DONE)
  {
    MMRESULT MMResult;
    WAVEHDR *WaveHeader = (WAVEHDR *)dwParam1;

    if((MMResult = waveOutUnprepareHeader(hwo, WaveHeader, sizeof(WAVEHDR))) != MMSYSERR_NOERROR)
    {
      char Text[256];
      waveOutGetErrorText(MMResult, Text, sizeof(Text));
      printf("Failed to unprepare: %s\n", Text);
    }

    free(WaveHeader->lpData);
  }
}

bool PlayMP3(const char* aBuffer, int aSize)
{
   HWAVEOUT WaveHandle;
   WAVEHDR WaveHeader;
   MMRESULT MMResult;

   MPEGLAYER3WAVEFORMAT Mp3Format;
   Mp3Format.wfx.cbSize = MPEGLAYER3_WFX_EXTRA_BYTES;
   Mp3Format.wfx.wFormatTag = WAVE_FORMAT_MPEGLAYER3;
   Mp3Format.wfx.nChannels = 1;
   Mp3Format.wfx.nAvgBytesPerSec = 24 * 1024/8;
   Mp3Format.wfx.wBitsPerSample = 0;
   Mp3Format.wfx.nBlockAlign = 1;
   Mp3Format.wfx.nSamplesPerSec = SAMPLE_RATE;
   Mp3Format.fdwFlags = MPEGLAYER3_FLAG_PADDING_OFF;
   Mp3Format.nBlockSize = 144 * 24*1024/SAMPLE_RATE; //MP3_BLOCK_SIZE;
   Mp3Format.nFramesPerBlock = 1;
   Mp3Format.nCodecDelay = 1393;
   Mp3Format.wID = 1;

   if(waveOutOpen(&WaveHandle, WAVE_MAPPER, (WAVEFORMATEX *)&Mp3Format, (DWORD )CallBackProc, 0, CALLBACK_FUNCTION) != MMSYSERR_NOERROR)
   {
      std::cerr << "Failed to open Wave!" << std::endl;
      return false;
   }
   memset(&WaveHeader, 0, sizeof(WaveHeader));
   aBuffer += 4; // Skip header for now
   aSize -= 4;
   WaveHeader.lpData = (char *)aBuffer;
   WaveHeader.dwBufferLength = aSize;
   WaveHeader.dwBytesRecorded = aSize;
   WaveHeader.dwLoops = 1;
   WaveHeader.dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;

   if(waveOutPrepareHeader(WaveHandle, &WaveHeader, sizeof(WaveHeader)) != MMSYSERR_NOERROR)
   {
     printf("Failed to prepare header!\n");
     return false;
   }

   if((MMResult = waveOutWrite(WaveHandle, &WaveHeader, sizeof(WaveHeader))) != MMSYSERR_NOERROR)
   {
     char Text[256];
     waveOutGetErrorText(MMResult, Text, sizeof(Text));
     printf("Failed to write: %s\n", Text);
     return false;
   }
   while(!(WaveHeader.dwFlags & WHDR_DONE))
     Sleep(100);
   if((MMResult = waveOutClose(WaveHandle)) != MMSYSERR_NOERROR)
   {
     char Text[256];
     waveOutGetErrorText(MMResult, Text, sizeof(Text));
     printf("Failed to close: %s\n", Text);
     return false;
   }
   return true;
}

int main(void)
{
   HRSRC ResHandle = FindResource(0, MAKEINTRESOURCE(1025), RT_RCDATA);
   if(!ResHandle)
   {
      std::cerr << "Failed to find resource" << std::endl;
      return 1;
   }

   HGLOBAL Res = LoadResource(0, ResHandle);
   if(!Res)
   {
      std::cerr << "Failed to load resource" << std::endl;
      return 1;
   }
   void *Ptr = LockResource(Res);
   if(!Ptr)
   {
      std::cerr << "Failed to lock resource" << std::endl;
      return 1;
   }
   DWORD Size = SizeofResource(0, ResHandle);

   PlayMP3((const char *)Ptr, Size);
   return 0;
}
